// -*- C++ -*-

// vchat.cc	Simple chat client

// $Header: /home/cvsroot/vchat/client/vchat.cc,v 1.11 1998/10/21 21:17:43 hans Exp $

// $Log: vchat.cc,v $
// Revision 1.11  1998/10/21 21:17:43  hans
// No longer send .t after login.  .s shows this.
//
// Revision 1.10  1998/10/21 21:13:58  hans
// Idle beep implementiert.
// Mit dem Client-Befehl %i kann der idle-beep-timer auf eine beliebige
// Anzahl von Sekunden gesetzt werden.  Der Client piept dann, wenn der
// User eine Nachricht eingegeben hat, und die naechste Nachricht, die
// vom Chatserver kommt, laenger als die eingestellte Zeit nach der
// letzten eigenen Eingabe kam.  Der Piep wird als zwei BELs mit einem
// Abstand von 0,4 Sekunden erzeugt, und ist dadurch auf manchen
// Mehrschirmkonsolen gut zu identifizieren.
//
// Man setzt den Timer sinnvollerweise auf ca. 5 Sekunden, und kann dann
// die Konsole mit dem Chat wegschalten.  Wenn dann wieder jemand was
// sagt, kriegt man es mit, ohne dass man permanent nachgucken muss.
//
// Revision 1.9  1998/07/31 23:07:46  hans
// Add userInput routine in vchat.pl.  This routine is being called to
// process the user's input before it is sent to the server.
//
// Revision 1.8  1998/07/04 16:38:17  hans
// Send .t command after log in.
//
// Revision 1.7  1998/07/04 00:13:41  hans
// Send .s command directly after log in.
//
// Revision 1.6  1998/05/30 22:29:30  hans
// Implement word wrapping of long lines.
//
// Revision 1.5  1998/04/13 21:40:39  hans
// moving to cvs
//
// Revision 1.4  1998/03/08 21:23:18  hans
// Perl added.  All server messages are now processed by the perl script
// defined by the environment variable CHATHELPERL.  The script must
// define a subroutine 'serverMessage' taking one string argument
// containing the server messages to be parsed.  It should return the
// formatted messages.  This can be used, for example, to translate the
// server messages into different languages.  In the future, vchat's API
// could be extended to allow for richer interactions with the perl
// backend.
//
// Revision 1.3  1998/03/07 22:01:11  hans
// Structural changes.
//

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>

#include <iostream.h>
#include <strstream.h>

#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <netdb.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>

#include <netinet/in.h>

#include "Terminal.h"
#include "TerminalCapability.h"
#if defined(perl)
#include "PerlHelper.h"
#endif

#define DEFAULT_SERVER "localhost:2323";


// termcap variables
Terminal *terminal;

// global quit flag - set to true if termination of main loop is desired
bool quit = false;

#if defined(perl)
PerlHelper *perlHelper;
#endif

// wrap 	Wrap long line into multiple shorter line.  Continuation
//		lines are indented by four spaces.

// This function is dedicated to Tim, who requested it in May 98.
// It is believed have (non-fatal) bugs.

static void 
wrapTalk(const char *const str, ostream &os)
{
  unsigned width = terminal->columns();
  const unsigned indent = 4;
  if (width <= (indent * 2)) {
    // cerr << "|terminal too narrow|";
    os.write(str, strlen(str));
    return;	// fail.  no space to wrap
  }

  const char *left = str;
  bool firstLine = true;	// width is lowered by indent after
  				// the first line has been printed
  for (;;) {
    // skip leading spaces
    while (*left && *left == ' ') {
      left++;
    }
    // terminate if string left to print shorter than width
    if (width >= strlen(left)) {
      // cerr << "|printing rest|";
      os.write(left, strlen(left));
      // cerr << endl;
      break;
    }
    // set up right to point to the last character to print
    const char *right = left + width;
    // scan for space until beginning of string
    while (right > left) {
      if (*right-- == ' ') {
	break;
      }
    }
    if (right > left) {
      // space found -> print string up to (not including) the found space
      // cerr << "|wrapping at " << right - left << "|";
      right++;	// adjust to point to space
      os.write(left, right - left);
      // and continue at the first character after the space
      left = right + 1;
    } else {
      // no space found -> cut hard
      // cerr << "|no space found, cutting hard" << "|";
      os.write(left, width);
      left += width;
    }

    os << "\r\n";

    if (firstLine) {
      // first line is not indented.
      // cerr << "|adjust|";
      width -= indent;
      firstLine = false;
    }
    // spit out indentation
    // cerr << "|indent|";
    for (int i=0; i<indent; i++) {
      os << ' ';
    }
  } 
}   

// server socket descriptor
int server;

void
processInput(const char *buf)
{
  if (terminal->cap("change_scroll_region")) {
    terminal->put(buf, terminal->cap("enter_bold_mode"));
  }

  ostrstream os;
  os << buf << "\r\n";

  if (::write(server, os.str(), os.pcount()) != os.pcount()) {
    cerr << "could not write to server" << endl;
    quit = true;
  }
  delete [] os.str();
}

void
interpretServerData(const char *buf)
{
#if defined(perl)
  const char *retval;
  //  cerr << "sending \"" << buf << "\" to perl" << endl;
  perlHelper->serverMessage(buf, &retval);
  //  cerr << "perl returned \"" << retval << "\"" << endl;
  istrstream is(retval);
#else
  istrstream is(buf);
#endif
  while (!is.eof()) {
    char code[4];
    char delim1, delim2;
    char *message;
    char c = is.peek();
    int isServerMessage = 0;
    code[0] = 0;
    if (isdigit(c)) {
      is.read(code, 3);
      code[3] = 0;
      is.read(&delim1, 1);
      isServerMessage = 1;
    }
    is.gets(&message, '\r');
    is.read(&delim2, 1);
    if (is.good()) {
      ostrstream os;
      if (isServerMessage) {
	os << "* " << message << '\0';
      } else {
	wrapTalk(message, os);
	os << '\0';
      }
      terminal->put(os.str());
      delete [] os.str();
    }
    delete [] message;
  }
#if defined(perl)
  delete [] retval;
#endif
}

int
eventHook(...)
{
  static char buf[16384];
  static count = 0;
  int len = ::read(server, buf + count, sizeof(buf) - count);
  //  cerr << "read " << len << " bytes" << endl;
  if (len <= 0) {
    if (len < 0 && errno == EAGAIN) {
      return 0;
    }
    ostrstream os;
    if (len < 0) {
      os << "* Server closed connection (" << strerror(errno) << ")" << '\0';
    } else {
      os << "* Server closed connection" << '\0';
    }
    terminal->put(os.str());
    delete [] os.str();
    quit = true;
    terminal->eventHook(0);
    return 0;
  }

  buf[len + count] = 0;
  if (buf[len + count - 1] == '\n') {
    interpretServerData(buf);
    count = 0;
  } else {
    count += len;
    if (count >= sizeof(buf)) {
      terminal->put("* Server sent bad message (no trailing nl) - exiting");
      quit = true;
      terminal->eventHook(0);
      return 0;
    }
  }

  return 0;
}

int
connectToServer(const char *serverName, short portNumber, const char *nick)
{
  {
    ostrstream os;
    os << "Connecting to chat server on host " << serverName
	 << " port " << portNumber << '\0';
    terminal->put(os.str());
    delete [] os.str();
  }

  server = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);

  if (server < 0) {
    ostrstream os;
    os << "Could not create socket (" << strerror(errno) << ")" << '\0';
    terminal->put(os.str());
    delete [] os.str();
    return 0;
  }

  struct hostent *he = gethostbyname(serverName);

  if (!he) {
    ostrstream os;
    os << "Could not resolve server host name " << serverName << '\0';
    terminal->put(os.str());
    delete [] os.str();
    return 0;
  }

  sockaddr_in serverAddress;
  memset((char *) &serverAddress, 0, sizeof(serverAddress));

  serverAddress.sin_family = AF_INET;
  serverAddress.sin_port = htons(portNumber);
  memcpy(&serverAddress.sin_addr, he->h_addr, sizeof(struct in_addr));
  
  if (connect(server,
	      (struct sockaddr *) &serverAddress,
	      sizeof(serverAddress)) < 0) {

    ostrstream os;
    os << "Could not connect to remote: " << strerror(errno) << '\0';
    terminal->put(os.str());
    delete [] os.str();
    return 0;
  }

  int flags = fcntl(server, F_GETFL, 0);
  fcntl(server, F_SETFL, flags | FNDELAY);

  return 1;
}

main(int argc, char *argv[])
{
  char *nick = getenv("CHATNICK");
  char *terminalType = getenv("TERM");
  char *from = getenv("FROM");
  char *chatserver = getenv("CHATSERVER");

  terminal = new Terminal(terminalType);

#if defined(perl)
  char *helperScript = getenv("CHATHELPERL");

  if (!helperScript) {
    helperScript = HELPER_SCRIPT;
  }

  perlHelper = new PerlHelper(helperScript);
#endif

  if (!chatserver) {
    chatserver = DEFAULT_SERVER;
  }

  istrstream is(chatserver);

  char *chatHost;
  short chatPort;
  is.gets(&chatHost, ':');
  is >> chatPort;

  if (!is.good()) {
    cerr << "bad chat server specification: " << chatserver << endl;
    exit(1);
  }
  
  while (!nick) {
    nick = terminal->readline("Enter nick name: ");
    if (!*nick) {
      free(nick);
      nick = 0;
    }
  }

  if (!from) {		// set up 'from' specification

    ostrstream os;
    char hostname[MAXHOSTNAMELEN+1]; hostname[MAXHOSTNAMELEN] = 0;
    gethostname(hostname, MAXHOSTNAMELEN);
    char domainname[MAXHOSTNAMELEN+1]; domainname[MAXHOSTNAMELEN] = 0;
    getdomainname(domainname, MAXHOSTNAMELEN);
    os << getlogin() << '@' << hostname;
    if (*domainname) {
      os << '.' << domainname;
    }
    os << '\0';

    from = os.str();
  }

  if (connectToServer(chatHost, chatPort, nick)) {

    ostrstream os;
    os << ".l " << nick << " " << (from ? from : "somewhere") << "\n";
    os << ".s" << "\n";
    write(server, os.str(), os.pcount());
    delete [] os.str();

    terminal->eventHook(eventHook);
    //  rl_redisplay_function = drawStatusLine;
    while (!quit) {
      char *buf = terminal->readline("");
#if defined(perl)
      const char *retval;
      //  cerr << "sending \"" << buf << "\" to perl" << endl;
      perlHelper->userInput(buf, &retval);
      //  cerr << "perl returned \"" << retval << "\"" << endl;
      if (*retval) {
	processInput(retval);
	delete [] retval;
      }
#else
      if (*buf) {
	processInput(buf);
      }
#endif
      free(buf);
      (*terminal) << terminal->cap("clr_eol");
      terminal->flush();
    }
  }

#if defined(perl)
  delete perlHelper;
#endif

  delete terminal;
  terminal = 0;

  cerr << endl << "Good bye." << endl << endl;
  sleep(1);
}

