#!/usr/local/bin/perl -w --

# This perl file is read by vchat.cc and contains helper functions
# called by vchat during the chat conversation.

# sub serverMessage

# Reformat one or more chat server messages.  The message(s) are
# passed in as one string in protocol format (see protocol.doc).  The
# routine should return the reformatted messages(s).  Note that
# vchat.cc will perform simple interpretations on every returned line
# beginning with three digits.  All other lines are put to the user's
# screen verbatim.

my $lastInput = time;
my $beeped = 1;
my $idleBeep = 0;

sub serverMessage {
    my $message = shift;

    my $retval = "";

    foreach (split(/\n/, $message)) {

	if (/^211 (\S+)/) {
	    $retval .= "* $1 joined the chat\r\n";

	} elsif (/^221 (\S+).*\((.*)\)/) {
	    my $reason = "";
	    if ($2 ne "") {
		$reason = " ($2)";
	    }
	    $retval .= "* $1 left the chat$reason\r\n";

	} elsif (/^241 (\S+) (\S+)/) {
	    $retval .= "* $1 changed his nick to $2\r\n";

	} elsif (/^118 (.*)/) {		# action
	    $retval .= "$1\r\n";

	} elsif (/^115 (\S+) *([^\r\n]*)/) {	# channel topic

	    my ($channel, $topic) = ($1, $2);

	    if ($topic eq "") {
		$retval .= "* No topic defined for channel $channel\r\n";
	    } else {
		$retval .= "* Topic for channel ${channel}: $topic\r\n";
	    }

	} elsif (/^116 (\S+) (\S+) (\S+) ([^\r\n]*)[\n\r]*$/) {
	    $retval .=
		"* Nickname: $1\r\n"
		    ."* From....: $2\r\n"
			."* Via.....: $3\r\n";
	    if ($4 ne "") {
		$retval .= "* Userinfo: $4\r\n";
	    }
	} else {
	    $retval .= $_ . "\n";
	}
    }

    if (($idleBeep > 0)
	and ((time - $lastInput) >= $idleBeep)
	and not $beeped) {
	$beeped = 1;
	print STDERR "\007";
	select(undef, undef, undef, 0.2);
	print STDERR "\007";
    }

    return $retval;
}

sub userInput {
    my $input = shift;

#    $input =~ s/<v*e*b*g>/<grin>/g;
    if ($input =~ /^%i\s*(\d)/) {
	$idleBeep = $1;
	$input = "";
    }

    $lastInput = time;
    $beeped = 0;

    return $input;
}
